""" Communication module.
=======================

Communicate with the different registries on the board.
Changing registries changes how the board and the chip operates.
This module gives the user low level control on how the board operates
through the 5 primary modules:
- analog registers
- digital registers
- control registers
- i2c
- serial

"""
from logging import getLogger

from .analog_registers import AnalogRegisters
from .control_registers import ControlRegisters
from .digital_registers import DigitalRegisters
from .i2c import *
from .i2c_registers import I2CRegisters

LOGGER = getLogger(__name__)


def debug_mode(board, activate=True):
    """Turn on debug on board.

    Normally the digital side of the ASIC handles the communication with
    the Analog portions but with debug enabled, all Analog commands
    are sent directly to the Analog part of the chip instead of to
    the digital part.

    This bypasses the digital controls entirely.

    Warning!
    Changes a ton of stuff in software and hardware so be careful
    Use at own risk!

    Args:
        active(bool): True enables debug mode, False disables
    """
    ControlRegisters(board).write("debug", activate)


def write_all_registers(board):
    """Updates all available registers to hardware.

    Will grab all the register variables stored in the board.params
    and send them to the hardware board.
    This function effectively sync the hardware with the software.

    Usefull to run on startup when the board is in a blank state or
    after running a program on the hardware that require the board to be
    returned to a known state.

    Args
        board (Board): Good ol' board object.
    """
    reg_types = [
        AnalogRegisters,
        ControlRegisters,
        DigitalRegisters,
        I2CRegisters,
    ]
    for reg in reg_types:
        try:
            reg(board).write_all()
        except (ValueError, TypeError, AttributeError) as error_msg:
            LOGGER.debug("Update registers failed due to : %s", error_msg)
        except KeyError as error_msg:
            LOGGER.debug("Can't update register due to KeyError: %s", error_msg)
        except Exception as error_msg:
            LOGGER.exception(
                "Update registers failed due to unknown error: %s", error_msg
            )



"""Collection of helper functions used in all modules.

"""
import datetime
import os
import socket
import sys
from typing import Iterable

import numpy as np
import numpy.typing as npt


def type_name(obj) -> str:
    """Get the name of an object's type.

    Args:
        obj (any): any object (classes are objects too!)

    Returns:
        The name as a string.
    """
    return type(obj).__name__


def swap_dict_key_and_val(to_swap: dict) -> dict:
    """Swap dict.

    Returns:
        {v, k} from {k, v}
    """
    return {v: k for k, v in to_swap.items()}


def hex_to_int(input_hex: str) -> int:
    """Converts hexdecimal string to an integer.

    Args:
        input_hex (str): hexdecimal string

    Returns:
        Integer
    """
    return int(input_hex, 16)


def int_to_bin(value, width=4):
    """Converts an integer to a binary string.

    Args:
        value (int): the value to convert to binary
        width (int): preferred width of the string.
            If the smallest possible binary representation of the
            value has fewer bits, the result is left-padded with
            zeros.

    Returns:
        _type_: _description_
    """
    if not isinstance(value, int):
        raise TypeError("Value must be an int")
    if not isinstance(width, int):
        raise TypeError("Width must be an int")
    if value < 0:
        raise ValueError("Negative values not supported")
    if width < 0:
        raise ValueError("Negative widths not supported")
    return f"{bin(value)[2:].zfill(width)}"


def locked_deque(it, lock):
    """Makes a process safe iterator from a deque.

    The locked deque can be safely used with the
    multiprocess map. This allows threads to write to a deque
    and a process pool to read from the deque to process it.

    Args:
        it (deque): object to iterate
        lock: the multiprocess lock.
    """
    while True:
        try:
            with lock:
                value = it.pop()
        except IndexError:
            return

        yield value


def get_application_path():
    """Application path depending on run mode:
        - frozen .exe
        - commandline
        - interactive mode (eg. notebook)

    Returns:
        Path to root of application.
    """
    if getattr(sys, "frozen", False):
        application_path = os.path.dirname(sys.executable)
    else:
        try:
            app_full_path = sys.modules["__main__"].__file__
            application_path = os.path.dirname(app_full_path)

        except (NameError, AttributeError):
            application_path = os.getcwd()

    return application_path


def create_dir_if_not_exist(directory):
    """Checks if a directory exists, creates it if it doesn't.

    Returns:
        The input directory but with the directory created.
    """
    if not os.path.exists(directory):
        os.makedirs(directory)
    return os.path.normpath(directory)


def get_package_versions() -> tuple:
    """Get the current naludaq and naluscope versions safely.

    Will not raise an exception if naluscope is not present.

    Returns:
        A tuple of (x, y) where x = naludaq version str, and
        y = naluscope version str or `None` if not installed.
    """
    try:
        # Ugly! ideally naludaq shouldn't care about naluscope
        import naluscope

        import naludaq

        return (naludaq.__version__, naluscope.__version__)
    except:
        return (naludaq.__version__, None)


def extract_event_creation_times(events: Iterable[dict]) -> npt.NDArray[np.datetime64]:
    """Extract event creation times from a collection of events into a numpy array.
    If creation times are not present, ``np.datetime64('nat')`` (not a time) is stored
    instead.

    Args:
        events (Iterable[dict]): events.

    Returns:
        np.ndarray[np.datetime64]: array of creation times with dtype of ``np.datetime64``.
    """
    try:
        iter(events)  # raises if not iterable
    except TypeError:
        raise

    from_timestamp = datetime.datetime.fromtimestamp
    times = np.full(len(events), np.datetime64("nat"), dtype="datetime64[us]")
    for idx, event in enumerate(events):
        if not isinstance(event, dict):
            raise TypeError(
                f"All events must be dicts. Got {type_name(event)} instead."
            )

        creation_time = event.get("created_at", None)
        if isinstance(creation_time, (int, float)):
            creation_time = from_timestamp(creation_time)
        if not isinstance(creation_time, (np.datetime64, datetime.datetime)):
            continue
        times[idx] = np.datetime64(creation_time)

    return times


def event_transfer_time(
    board,
    windows: int,
    channels: int = None,
    margin: float = 2.0,
    overhead: float = 0,
) -> float:
    """Calculate the approximate time it takes to transfer an event from the board.

    The event size is approximated based on the number of windows and channels,
    and does not account for headers/footers (although this is typically negligible).

    Args:
        board (Board): the board object. Must have valid connection info
        windows (int): number of windows per channel in the event
        channels (int): number of channels returned in the event. Defaults to all channels.
        margin (float): value to scale the transfer time by to adjust for real-world scenarios.
        overhead (float): overhead time in seconds time to add to adjust for real-world scenarios.

    Returns:
        float: the approximate time in seconds it would take to transfer such an event.
    """
    from naludaq.backend.managers import ConnectionManager

    if channels is None:
        channels = board.channels
    speed = board.connection_info.get("speed", 115_200)
    if board.using_new_backend:
        speed = 30_000_000
        cm = ConnectionManager(board)
        if cm.is_uart_based:
            speed = cm.device.baud_rate
    # 2 bytes per sample
    event_size = channels * windows * board.params.get("samples", 64) * 2
    transfer_rate = speed // 8
    transfer_time = event_size / transfer_rate
    adjusted_time = transfer_time * margin + overhead
    return adjusted_time


def get_available_port(host: str = "127.0.0.1") -> int:
    """Get an available port for the server to use.

    Args:
        host (str): the host to bind to. Defaults to localhost.

    Raises:
        ConnectionError: if no port is available.
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.bind((host, 0))
            return s.getsockname()[1]
    except:
        raise ConnectionError("Could not find an available port")



"""Gain stage controller for the aodsv2_eval board.
"""
from copy import deepcopy

from naludaq.communication import LOGGER
from naludaq.communication.analog_registers import AnalogRegisters
from naludaq.communication.control_registers import ControlRegisters

from .aodsv2 import GainStageController as AODSV2GainStageController


class GainStageController(AODSV2GainStageController):
    """Controls the AODS gain stages

    Gain stages are controlled by the following registers:
    - sel_0: ch0 input buffer
    - sel_1: ch1 input buffer
    - sel_2: ch2 input buffer
    - sel_3: ch3 input buffer
    - sel_4: external output
    - sel_5: ch1 BAMP or x8 of ch0 output
    - sel_6: select either BAMP or 8x gain for ch3
    - sel_7: select ch3 input either ch0 output or ch2 output

    There are several functions in the form `ch{X}_{some possible configuration}`
    such as `ch3_external_input()` that serve as presets and a human-readable
    way to configure each stage. Note that the configurations for each channel
    are mutually exclusive; e.g. you cannot have ch1 external input and ch1 8x ch0
    simulatenously.
    """

    def __init__(self, board, chip_number=0):
        """Initialize the controller"""
        super().__init__(board)
        self.chip_num = chip_number

        # Since we don't support multiple chips in the registers yet, we need to
        # keep track of the values of the registers ourselves.
        self._sel_values = deepcopy(
            board.registers["analog_registers"]
        )  # TODO: fix this when multichip registers are fixed

    @property
    def sel_values(self) -> "list[bool]":
        """Get a list of sel values from the internally stored registers"""
        return [
            self._sel_values[f"sel_{i}"][
                "value"
            ]  # TODO: fix this when multichip registers are fixed
            for i in range(self._num_stages)
        ]

    def _write_reg(self, reg, value):
        """Write a register.

        We need to override the chip mask manually since we don't support
        multichip registers yet.
        """
        LOGGER.debug("Selecting chip %d", self.chip_num)
        self._set_chip_tx_enabled(
            [self.chip_num]
        )  # TODO: fix this when multichip registers are fixed
        AnalogRegisters(self.board).write(reg, value)
        self._sel_values[reg][
            "value"
        ] = value  # TODO: fix this when multichip registers are fixed
        self._set_chip_tx_enabled(
            range(self.board.params.get("num_chips", 2))
        )  # TODO: fix this when multichip registers are fixed

    def _set_chip_tx_enabled(
        self, chips: list = [0, 1]
    ):  # TODO: fix this when multichip registers are fixed
        """Set whether the TX lines are enabled for each chip.
        Can be used to write a register on a single chip instead of both.
        """
        ControlRegisters(self.board).write("ard_tx_en", sum([1 << x for x in chips]))



import numpy as np


def process_peds(pedestals, filter_outlier=True, filter_threshold=10, channels=None):
    """Filters the outliers from data and rawdata and finds std values for rawdata.

    This function will remove outliers from a copy of pedestals unless specified, leaving the input
    pedestals untouched.

    Args:
        pedestals (dict): Pedestals with params "data" and "raw_data"
        filter_outlier (bool): True to filter outliers
        filter_threshold (float): The modified z-score used to determine if a point is an outlier.
        channels (list): The channels to process peds

    Returns:
        ped (dict):
            Processed pedestal data with keys: "data" & "std_data" and
            "num_removed" if filter_outlier is enabled, with channels in
            the order of the channels argument.
    """
    num_channels = pedestals["params"]["channels"]
    return_ped = {
        "data": [[] for _ in range(num_channels)],
        "std_data": [[] for _ in range(num_channels)],
    }

    if channels is None:
        channels = range(0, pedestals["params"]["channels"])
    if filter_outlier:
        return_ped["num_removed"] = {
            "data": [0] * num_channels,
            "std_data": [0] * num_channels,
        }

    for chan in channels:
        ped_data = np.hstack(pedestals["data"][chan])
        if pedestals["params"]["model"] == "hdsocv1":
            stitched_rawpeds = []
            lengths = [len(p) for p in pedestals["rawdata"][chan]]

            for i in range(min(lengths)):
                window = [p[i] for p in pedestals["rawdata"][chan]]
                stitched_rawpeds.append(np.hstack(window))

            ped_stddata = np.std(stitched_rawpeds, axis=0)
        else:
            ped_stddata = np.std(pedestals["rawdata"][chan], 2).flatten()
        if filter_outlier and not np.all(
            ped_data == ped_data[0]
        ):  # Filter outliers utilizing a modified z-score
            data_length = len(ped_data)
            stddata_length = len(ped_stddata)
            ped_data = filter_outliers(ped_data, filter_threshold)
            ped_stddata = filter_outliers(ped_stddata, filter_threshold)
            return_ped["num_removed"]["data"][chan] = data_length - len(ped_data)
            return_ped["num_removed"]["std_data"][chan] = stddata_length - len(
                ped_stddata
            )
        return_ped["data"][chan] = ped_data
        return_ped["std_data"][chan] = ped_stddata

    return return_ped


def is_outlier(data, thresh=3.5):
    """Using a modified Z-score to detect outliers.

    Returns a boolean array with True if points are outliers and False
    otherwise.

    Args:
        data (np.array): numpy array of data
        thresh (float) : The modified z-score to use as a threshold. Observations with
            a modified z-score (based on the median absolute deviation) greater
            than this value will be classified as outliers.

    Returns:
        A numobservations-length boolean array.

    References:
        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
    """
    if len(data.shape) == 1:
        data = data[:, None]
    median = np.median(data)
    diff = np.sum((data - median) ** 2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)
    modified_z_score = 0.6745 * diff / med_abs_deviation

    return modified_z_score > thresh


def filter_outliers(data, thresh):
    """Filter outliers from the data and return data without outliers.

    This function doesn't give an indication of where the outliers are, it just removes them.
    Useful to run before plotting data to remove outliers.

    Args:
        data (np.array): data array to filter
        thresh: Z-value threshold, see `is_outlier` function description.

    Returns:
        filtered np.array
    """

    indexes = np.where(is_outlier(data, thresh) == False)  # invert

    modified_data = data[indexes]

    return modified_data



from http import HTTPStatus


class HttpError(Exception):
    def __init__(self, status_code: HTTPStatus, response: dict, *args: object):
        """Exception for HTTP errors received from the backend.

        Args:
            status_code (HTTPStatus): the error status code
            response (dict): the response JSON sent by the backend
        """
        super().__init__(*args)
        self._status_code = status_code
        self._error_id = response["error_id"]
        self._message = response["message"]

    @property
    def status_code(self) -> HTTPStatus:
        """Get the error status code"""
        return self._status_code

    @property
    def error_id(self) -> int:
        """Get the error ID"""
        return self._error_id

    @property
    def message(self) -> str:
        """Get the error message"""
        return self._message

    def raise_typed(self):
        """Raise the error as a specific type depending on the error ID"""
        type = {
            0: BackendError,
            1: ValueError,
            2: ConnectionError,
            3: DeviceError,
            4: ConnectionError,
            5: ConnectionError,
            6: ConnectionError,
            7: AcquisitionError,
            8: ValueError,
            9: ValueError,
            10: IndexError,
            11: AcquisitionError,
            12: BackendError,
            13: TimeoutError,
        }.get(self.error_id, BackendError)
        message = f"[Status {self.status_code}, ID {self.error_id}] {self.message}"
        exception = type(message)
        raise exception


class BackendError(Exception):
    pass


class BackendUnreachableError(BackendError):
    pass


class ConnectionError(BackendError):
    pass


class AcquisitionError(BackendError):
    pass


class ContextError(BackendError):
    pass


class DeviceError(BackendError):
    pass



"""
Driver for talking over a FT60X interface interface

Should be interoperable with the linketh package


Marcus Luck - Nalu Scientific
email: marcus@naluscientific.com
January 2023
"""
import binascii
import ctypes
import logging
import platform

try:
    import ftd3xx as ftd
except (FileNotFoundError, OSError):
    raise FileNotFoundError("D3XX driver not installed")
except ImportError:
    raise ImportError("ftd3xx not installed. Please install package.")

if platform.system() == "Windows":
    import ftd3xx._ftd3xx_win32 as _ft
# elif platform.system() == "Linux":
#     import ftd3xx._ftd3xx_linux as _ft
else:
    raise ImportError(f"Support for USB 3 on {platform.system()} is unavailable")

import numpy as np

from naludaq.board.connections.base_serial import BaseConnection
from naludaq.helpers.exceptions import FTDIError
from naludaq.parsers.answer_parser import get_answer_parser

logger = logging.getLogger("naludaq.D3XX")
_WRITE_MESSAGE_NIBBLES = 8
_DEFAULT_READ_PIPE_ID = 0x82
_DEFAULT_WRITE_PIPE_ID = 0x02
_DEFAULT_CHUNK_SIZE = 4 * 1024  # 786_842


class USB3(BaseConnection):
    """FTDI is a communications class for talking to the boards.

    Uses the ftd2xx Python package, which requires d2xx drivers
    to be properly installed. For information about using d2xx:
    https://ftdichip.com/wp-content/uploads/2020/08/D2XX_Programmers_GuideFT_000071.pdf
    """

    def __init__(self, conn_info: dict):
        """Initializes the FTDI object, but does not open a connection.

        Args:
            conn_info (dict): the connection info dict. Required keys: 'speed', 'model',
                and 'serial_number'. 'stop_word' is preferred, and defaults
                to 0xCAFE if not present.

        Raises:
            ConnectionError if the connection info dict is missing the required keys.
        """
        super().__init__()
        self._conn: ftd.FTD3XX = None
        self._read_addr = "AD"
        self._write_addr = "AF"
        self._is_open = False

        self._chunk_size = _DEFAULT_CHUNK_SIZE
        self._read_pipe_id = _DEFAULT_READ_PIPE_ID
        self._write_pipe_id = _DEFAULT_WRITE_PIPE_ID

        self._speed = conn_info["speed"]
        self._model = conn_info["model"]
        self.stopword = conn_info.get("stop_word", b"\xca\xfe")  # b'\xfa\xce'
        self.new_firmware = conn_info.get("new_firmware", False)
        self.response_length = 4
        self.parse_answer = get_answer_parser()

        serial_number = conn_info.get("serial_number", None)
        if serial_number:
            logger.debug("using serial_number %s to connect", serial_number)
            self.serial_number = serial_number
        else:
            raise ConnectionError("FTDI connection requires the serial num or port num")

    def __del__(self):
        """Closes the connection if one is open."""
        self.close()

    @property
    def new_firmware(self):
        """Get/Set if the board uses a new firmware or not.

        If a new firmware is used, the register answers are different.
        """
        return self._new_firmware

    @property
    def in_waiting(self):
        return self._chunk_size

    @new_firmware.setter
    def new_firmware(self, value: bool):
        self._new_firmware = value
        self.response_length = 4
        if value is True:
            self.response_length = 8
        self.parse_answer = get_answer_parser(value)

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        is_open = True
        try:
            self._conn.getChipConfiguration()
        except (AttributeError, ftd.DeviceError):
            is_open = False
        return is_open

    def open(self):
        """Open connection over the FTDI interface.

        Set board (device) number and baud before opening a connection, use:
            .serial_number = value
        """
        if self.is_open:
            logger.debug("USB3.open(): device already open")
            return

        serial_number = self.serial_number.encode("ascii")
        usb_connection = ftd.create(serial_number, _ft.FT_OPEN_BY_SERIAL_NUMBER)
        self._conn = usb_connection
        self._is_open = usb_connection is not None

        if usb_connection is None:
            raise ConnectionError("Failed to connect USB3")

    def close(self):
        """Close the serial port."""
        if self._conn is not None:
            try:
                self._conn.close()
            except Exception as e:
                logger.debug("Failed to close connection due to: %s", e)
            self._conn = None
        self._is_open = False

    def send(self, data, pause=0.02):
        """Converts the string to a series of bytes to send in the correct format

        Input:
            data (hex): A hex type string (interface needs 32 bits, so 8 hex chars minimum)
            pause (float): How logn to wait in between send.
        """
        if len(data) % _WRITE_MESSAGE_NIBBLES != 0:
            logger.debug(
                "FTDI.send(): need multiples of 8 hex chars, you gave me %s", len(data)
            )
            logger.debug("command:%s", data)
            return

        try:
            tosend = binascii.a2b_hex(data)
        except binascii.Error as e:
            logger.error(f"FTDI.send(): could not convert hex data: {e}")
            return

        logger.debug("Sending: %s", data)
        try:
            bytes_written = self._conn.writePipe(
                self._write_pipe_id, tosend, len(tosend)
            )
        except AttributeError as error_msg:
            logger.error(
                "Can't send command to board, no connection available. %s", error_msg
            )
        if bytes_written == 0:
            raise ConnectionError("Failed to send data, no connection available")

    def read_until(self, stopword: bytes) -> bytearray:
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        lenstop = len(stopword)
        buff = bytearray()
        while True:
            value = None
            try:
                value = self.read(self._chunk_size)
            except Exception as e:
                logger.debug("Receiving data failed due to: %s", e)
            if value is None or len(value) == 0:
                logger.error("Failed to read data")
                break

            buff += value
            if buff[-lenstop:] == stopword:
                break

        logger.debug("Received: %s", buff.hex())

        return buff

    def receive(self):
        """Reads out an event or answer. Both end with a stopword.

        This will *NOT* work on boards where data and registers have different stopwords.
        """
        return self.read_until(self.stopword)

    def read(self, length: int = None) -> bytes:
        """Attempt to read data from the connection.

        Args:
            length (int, optional): maximum number of bytes to read.
                If not provided, the chunk size is assumed.

        Raises:
            FTDIError: If the operation failed.

        Returns:
            bytes: The data received. This may be less than the amount requested.
        """
        if length is None:
            length = self._chunk_size
        try:
            data = ctypes.c_buffer(length)
            rxlength = self._conn.readPipe(self._read_pipe_id, data, length)
            data = data[0:rxlength]
        except ftd.DeviceError as e:
            raise FTDIError(f"USB.read failed due to: {e}")
        if data == b"\xFA\xCE":
            return b""
        return data

    def read_all(self, attempts=10) -> bytearray:
        """Read all waiting bytes"""
        data = bytearray()

        attempt_count = 0
        while attempt_count < attempts:
            resp = self.read(self._chunk_size)
            if len(resp) == 0:
                attempt_count += 1
                continue
            data += resp
        return data

    def reset_input_buffer(self):
        """Discards any data currently in the input buffer."""
        self.read_all()

    def reset_output_buffer(self):
        """Not implemented for USB3."""

    def set_answer_parser(self, old=True):
        """Set's the answer parser"""
        self.parse_answer = get_answer_parser(old)

    def readReg(self, regNum, timeout=10000) -> int:
        """
        Reads out the register specified by the integer regNum
        Args:
            regNum (int): The registry number to read
            timeout (int): Amount of retries.
        Returns:
            Hex value of the serialread of two bytes.
        """
        # Send command
        self._send_readreg_cmd(regNum)

        # Wait for response
        buffer = self.receive()
        # Parse response
        answer = self.parse_answer(buffer)
        # Make sure responses are valid

        try:
            logger.debug(
                "FTDI.readReg(%s) success: %s %s %s",
                answer["read_reg"],
                answer["value"],
                hex(answer["value"]),
                "{0:b}".format(answer["value"]).zfill(16),
            )
        except Exception:
            pass

        return answer["value"]

    def _send_readreg_cmd(self, regNum):
        """Generates and sends a command to read a register.

        Args:
            regNum (str, int): address as int or hex str.
        """
        cmd_id = np.random.randint(0, 2**16)
        if isinstance(regNum, str):
            regNum = int(regNum, 16)
        read_addr = int(self._read_addr, 16)

        try:
            command = f"{read_addr:x}{regNum:02x}{cmd_id:04x}"
        except Exception as error_msg:
            logger.exception(f"send_readreg_cmd create cmd failed: {error_msg}")
        else:
            self.send(command)



"""Gain stage controller for the aodsv2_eval board.
"""
from logging import getLogger
from typing import List, Tuple

from naludaq.communication.analog_registers import AnalogRegisters
from naludaq.controllers.controller import Controller

LOGGER = getLogger("naludaq.gainstage_controller")


class GainStageController(Controller):
    """Controls the AODS gain stages

    Gain stages are controlled by the following registers:
    - sel_0: ch0 input buffer
    - sel_1: ch1 input buffer
    - sel_2: ch2 input buffer
    - sel_3: ch3 input buffer
    - sel_4: external output
    - sel_5: ch1 BAMP or x8 of ch0 output
    - sel_6: select either BAMP or 8x gain for ch3
    - sel_7: select ch3 input either ch0 output or ch2 output

    There are several functions in the form `ch{X}_{some possible configuration}`
    such as `ch3_external_input()` that serve as presets and a human-readable
    way to configure each stage. Note that the configurations for each channel
    are mutually exclusive; e.g. you cannot have ch1 external input and ch1 8x ch0
    simulatenously.
    """

    def __init__(self, board):
        """Initialize the controller"""
        super().__init__(board)
        self._num_stages = 8

    @property
    def sel_values(self) -> "list[bool]":
        """Get a list of sel values from the board registers"""
        analog_registers = self.board.registers["analog_registers"]
        return [analog_registers[f"sel_{i}"]["value"] for i in range(self._num_stages)]

    def set_manual(self, stages: List[bool]) -> List[Tuple[int, int]]:
        """Manually set the gainstages.

        This will set the gainstages to the values of the index in the list
        Each index corresponds to a gainstage and can be either True or False.

        Args:
            stages (List[bool]): List of boolean values.

        Returns:
            List[Tuple[int, int]]: List of tuples with (channel input, total gain)

        Raises:
            ValueError: If the length of the gainstages list is not 7.
        """
        self._validate_manual_input_or_raise(stages)
        for i, stage in enumerate(stages):
            self._write_reg(f"sel_{i}", stage)
        return self.compute_gains()

    def _validate_manual_input_or_raise(self, stages: List[bool]):
        """Validate the manual stages and raise an error if invalid"""
        if not isinstance(stages, list):
            raise TypeError("stages must be a list")
        if not all([s in [0, 1] for s in stages]):
            raise TypeError("Expected a list of boolean values")
        if len(stages) != self._num_stages:
            raise ValueError(f"Expected {self._num_stages} stages, got {len(stages)}")

    def ch0_external_input(self):
        """Disable ch0 input buffer"""
        sel_values = self.sel_values
        sel_values[0] = 0
        return self.set_manual(sel_values)

    def ch0_external_input_buffered(self):
        """Enable ch0 buffered input"""
        sel_values = self.sel_values
        sel_values[0] = 1
        return self.set_manual(sel_values)

    def ch1_external_input(self):
        """Disable ch1 input buffer"""
        sel_values = self.sel_values
        sel_values[1] = 0
        sel_values[5] = 0  # don't care
        return self.set_manual(sel_values)

    def ch1_512x_ch0_input(self):
        """Enable ch1 BAMP of ch0 output"""
        sel_values = self.sel_values
        sel_values[1] = 1
        sel_values[5] = 1
        return self.set_manual(sel_values)

    def ch1_8x_ch0(self):
        """Enable ch1 8x of ch0 output"""
        sel_values = self.sel_values
        sel_values[1] = 1
        sel_values[5] = 0
        return self.set_manual(sel_values)

    def ch2_external_input(self):
        """Enable ch2 external input"""
        sel_values = self.sel_values
        sel_values[2] = 0
        return self.set_manual(sel_values)

    def ch2_8x_ch1(self):
        """Enable ch2 8x input"""
        sel_values = self.sel_values
        sel_values[2] = 1
        return self.set_manual(sel_values)

    def ch3_external_input(self):
        """Enable ch3 external input"""
        sel_values = self.sel_values
        sel_values[3] = 0
        sel_values[6] = 0  # don't care
        sel_values[7] = 0  # don't care
        return self.set_manual(sel_values)

    def ch3_8x_ch0(self):
        """Enable ch3 8x input"""
        sel_values = self.sel_values
        sel_values[3] = 1
        sel_values[6] = 0
        sel_values[7] = 1
        return self.set_manual(sel_values)

    def ch3_512x_ch0(self):
        """Enable ch3 BAMP input"""
        sel_values = self.sel_values
        sel_values[3] = 1
        sel_values[6] = 1
        sel_values[7] = 1
        return self.set_manual(sel_values)

    def ch3_8x_ch2(self):
        """Enable ch3 8x of ch2 output"""
        sel_values = self.sel_values
        sel_values[3] = 1
        sel_values[6] = 0
        sel_values[7] = 0
        return self.set_manual(sel_values)

    def ch3_512x_ch2(self):
        """Enable ch3 BAMP of ch2 output"""
        sel_values = self.sel_values
        sel_values[3] = 1
        sel_values[6] = 1
        sel_values[7] = 0
        return self.set_manual(sel_values)

    def enable_external_output(self):
        """Enable external output"""
        sel_values = self.sel_values
        sel_values[4] = 1
        return self.set_manual(sel_values)

    def disable_external_output(self):
        """Disable external output"""
        sel_values = self.sel_values
        sel_values[4] = 0
        return self.set_manual(sel_values)

    def compute_gains(self, sel: "list[bool]" = None):
        """Compute the gains for the current configuration"""
        if sel is None:
            sel = self.sel_values
        self._validate_manual_input_or_raise(sel)

        prev_ch = []
        prev_ch.append((0, 1))  # ch0
        prev_ch.append(self._compute_ch1(sel))  # ch1
        prev_ch.append(self._compute_ch2(sel, prev_ch))  # ch2
        prev_ch.append(self._compute_ch3(sel, prev_ch))  # ch3
        prev_ch.append(self._compute_ext(sel, prev_ch))  # ext

        return prev_ch

    def _compute_ch1(self, sel) -> Tuple[int, int]:
        """Return a tuple with (channel input, total gain)"""
        ch_input = 1
        gain = 1
        if sel[1]:
            ch_input = 0
            gain = -8
            if sel[5]:
                gain = -512

        return (ch_input, gain)

    def _compute_ch2(
        self, sel, prev_stages: "list[tuple[int, int]]"
    ) -> Tuple[int, int]:
        """Return a tuple with (channel input, total gain)"""
        prev_stage = prev_stages[-1]
        ch_input = 2
        gain = 1
        if sel[2]:
            ch_input = prev_stage[0]
            gain = -8 * prev_stage[1]

        return (ch_input, gain)

    def _compute_ch3(
        self, sel: "list[int]", prev_stages: "list[tuple[int, int]]"
    ) -> Tuple[int, int]:
        """Return a tuple with (channel input, total gain)"""
        prev_stage = (0, 1) if sel[7] else prev_stages[-1]
        prev_gain = prev_stage[1]

        ch_input = 3
        gain = 1
        if sel[3]:
            ch_input = prev_stage[0]
            gain = -8 * prev_gain
            if sel[6]:
                gain = -512 * prev_gain

        return (ch_input, gain)

    def _compute_ext(self, sel, prev_stages) -> Tuple[int, int]:
        """Return a tuple with (channel input, total gain)"""
        ch_input, gain = prev_stages[3]
        if not sel[4]:
            ch_input = prev_stages[3][0]
            gain = 0

        return (ch_input, gain)

    def _write_reg(self, reg, value):
        """Write a register"""
        AnalogRegisters(self.board).write(reg, value)

    def __repr__(self) -> str:
        """Return a string representation of the object"""
        gains = self.compute_gains()
        return "".join(
            [
                f"self.__class__.__name__ - ",
                f"ch0=ch{gains[0][0]}x{gains[0][1]}), ",
                f"ch1=ch{gains[1][0]}x{gains[1][1]}), ",
                f"ch2=ch{gains[2][0]}x{gains[2][1]}), ",
                f"ch3=ch{gains[3][0]}x{gains[3][1]}), ",
                f"ext=ch{gains[4][0]}x{gains[4][1]})",
            ]
        )



def get_pedestals_generator_new(
    board,
    num_captures: int = 10,
    num_warmup_events: int = 10,
    channels: "list[int]" = None,
):
    if board.model in ["upac32", "upaci", "zdigitizer"]:
        raise NotImplementedError("Board is currently unsupported")
    elif board.model in [
        "aardvarcv3",
        "aardvarcv4",
        "trbhm",
        "aodsoc_aods",
        "aodsoc_asoc",
    ]:
        from .aav3 import PedestalsGeneratorAardvarcv3New

        return PedestalsGeneratorAardvarcv3New(
            board, num_captures, num_warmup_events, channels
        )
    elif board.model in ["hdsocv1_evalr2"]:
        from .hdsoc import PedestalsGeneratorHdsocNew

        return PedestalsGeneratorHdsocNew(
            board, num_captures, num_warmup_events, channels
        )
    elif board.model in ["udc16"]:
        raise NotImplementedError("Board is currently unsupported")
    elif board.model in ["upac96"]:
        raise NotImplementedError("Board is currently unsupported")
    else:
        from .default import PedestalsGeneratorNew

        return PedestalsGeneratorNew(board, num_captures, num_warmup_events, channels)



"""ASoCv3Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging

import numpy as np

from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class ASoCv3Parser(Parser):
    def __init__(self, params):
        super().__init__(params)

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.
        Assuming that the data packets are constant length, we can extract the data
        in place with matrix operations, speeding up the parsing.

        Parses the data spit out by the digitial readout portion of the firmware

        Channel data is interleaved, e.g
        pkt0,4,8,.... - ch0,
        pkt1,5,9,.... - ch1,
        etc

        TODO: To speed this up more, we do not want two matrices (bitshift & raw),
        we would perferably only want one.

        Args:
            in_data (bytearray): Raw data from the board
        Returns:
            Parsed event as a dict.
        Raises:
            BadDataException if no data is found or if the data contains errors.

        Additional info:
        The word is 16-bit and the data is 12-bit.
        The data is in the form:
        ZZZX XXXX XXXX XXXZ
        Where Z are unused bits.
        0x00 is not a valid byte, ever

        Example format for ASoC:
        first three bits are the "identifier", last bit is always 1
        4 x event headers:
        '001' 'prev_final_window' '1'
        '001' 'trigger_time_ns(23 dt 12)' '1'
        '001' 'trigger_time_ns(11 dt 0 )' '1'
        '001' 'eventNumber' '1'
        1x window header:
        '010' 'chan' '00' 'window' '1'
        64x data packets
        '100' 'data' '1'
        1x end packet (x"FACE")

        """
        try:
            self._validate_input_package(in_data)
        except:
            raise
        # Convert the 8-bit binary to 16-bit unsigned int and bit mask+shift the entire array
        raw_data = np.array(
            np.frombuffer(in_data["rawdata"], dtype=">H"), dtype="uint16"
        )

        # The raw_data is in the form:
        # ZZZX XXXX XXXX XXXZ
        # Where Z are unused bits.

        # Bitshifted data will make it in the form:
        # 0000 XXXX XXXX XXXX
        bitshifted_data = self._bitshift_data(raw_data)

        curr_event = {
            "window_labels": [[] for _ in range(self.params["channels"])],
            "data": [[] for _ in range(self.params["channels"])],
            "start_window": np.uint16(
                (raw_data[self.params["headers"]] & self.params["windmask"])
                >> self.last_bits
            ),
        }

        self._parse_event_headers(curr_event, bitshifted_data)

        # Reshape the 1D array into a 2D array, packet major uint minor
        num_headers = self.params["headers"]
        data_end = len(bitshifted_data) - self.num_footers

        event_size = len(bitshifted_data) - self.num_footers - num_headers + 1
        packet_size = self.num_wind_header + self.params["samples"]
        num_packet = event_size // packet_size

        event_matrix = np.reshape(
            bitshifted_data[num_headers : data_end + 1],
            [packet_size, num_packet],
            order="F",
        )
        raw_event_matrix = np.reshape(
            raw_data[num_headers : data_end + 1], [packet_size, num_packet], order="F"
        )
        # Pull channel info from all packets
        channels = (raw_event_matrix[0, :] & self.params["chanmask"]) >> self.params[
            "chanshift"
        ]
        available_channels = np.unique(channels[: self.params["channels"]])
        for idx, chan in enumerate(available_channels):
            curr_event["data"][chan] = (
                event_matrix[1:packet_size, idx :: len(available_channels)]
            ).flatten(order="F")
            curr_event["window_labels"][chan] = list(
                (
                    raw_event_matrix[0, idx :: len(available_channels)]
                    & self.params["windmask"]
                )
                >> self.last_bits
            )

        return curr_event



"""Readout settings for the aardvarcv3 board.

When setting the readout channel, aardvarc uses exclude channel mask registers
to exclude certain channels from readout

Controls the readout channels and the readout windows.
"""
from logging import getLogger

from .default import ReadoutController

LOGGER = getLogger("naludaq.readout_controller_aardvarcv3")


class Aardvarcv3ReadoutController(ReadoutController):
    def __init__(self, board):
        """Readout Controller for aardvarcv3.
        Args:
            board (Board): the board object.
        """
        super().__init__(board)

    def set_readout_channels(self, channels_to_read: list):
        """Select channels to readout.

        Update the registers and write them to the board.

        Args:
            channels_to_read(list): List of channel numbers to read

        Raises:
            TypeError if the list or an element in the list are the wrong type
            ValueError if the list is too large, too small, or contains a channel
                number that is out of bounds.

        """
        self._validate_channels_or_raise(channels_to_read)
        chansel = self._generate_channels_bits(channels_to_read)
        max_chan = self.board.params["channels"]
        # Exclude channel mask is opposite of chansel
        excludechanmask = (2**max_chan - 1) ^ int(chansel, 2)
        self._write_digital_register("excludechannelmask", excludechanmask)

    def get_readout_channels(self):
        """Get the current channels to read out.

        Returns:
            Sorted list of channel numbers to read out.
        """
        mask = self.board.registers["digital_registers"]["excludechannelmask"]["value"]
        return [c for c in range(self.board.channels) if (mask >> c) & 1 == 0]



from naludaq.communication import AnalogRegisters, DigitalRegisters
from naludaq.controllers.controller import Controller
from naludaq.helpers.semiton import SemitonABC


class BiasingModeControllerUDC16(Controller, SemitonABC):
    def __init__(self, board):
        """Controls how the internal amplifiers are generally biased.

        Current biasing is mode stable but cant be tweaked (unless reflow).
        Voltage biasing is less stable but can be tweaked

        NOT to be confused with external DAC biasing.
        Will need to retake pedestals when changing biasing mode.

        Args:
            board (Board): Board to change biasing mode
        """
        super().__init__(board)

    def set_voltage_biasing(self):
        """Quick access function to select which channels to operate in voltage biasing mode."""

        self.set_biasing_mode(chan0_7="v", chan8_15="v")

    def set_current_biasing(self):
        """Quick access function to select which channels to operate in current biasing mode."""

        self.set_biasing_mode(chan0_7="c", chan8_15="c")

    def set_biasing_mode(self, chan0_7: "str|None" = None, chan8_15: "str|None" = None):
        """Changes between either voltage or current biasing type

        Voltage biasing is the "traditional" way of biasing the input signal
        via external DAC. Current mode is using a fixed biasing that is supposed to
        be more stable, but is experimental for this design.

        Args:
            bias_mode (str): "current" or "voltage" mode selection
            chan0_7 (str, optional): if true, will change channels 0-7 to biasing mode, defaults to "True".
            chan8_15 (str, optional): if true, will change channels 0-7 to biasing mode, defaults to "True".
        """

        self._validate_inputs_or_raise(chan0_7)
        self._validate_inputs_or_raise(chan8_15)
        self._reset_side_selectors()

        biasing_registers = {
            "comp": (0, 0xFF),
            "ramp7_4": (0, 0xF),
            "ramp3_0": (0, 0xF),
            "transfer": (0, 0xFF),
        }

        # Registers to switch comparator, transfer buffer, and ramp to either current or voltage biasing
        # Logical 1 sets the component to voltage mode, logical 0 sets to current mode
        for bias_mode, side in zip([chan0_7, chan8_15], ["l", "r"]):
            if bias_mode is None:
                continue
            bias_mode = bias_mode.lower()[0]
            paramname = {"l": "0_7", "r": "8_15"}[side]
            if bias_mode != self.board.params[f"intamp_bias_mode_chan{paramname}"]:
                self._write_digital_register(f"writemask_common_{side}", 1)

                sel = {"c": 0, "v": 1}[bias_mode]

                for register, value in biasing_registers.items():
                    self._write_analog_register(register, value[sel])

                self._write_digital_register(f"writemask_common_{side}", 0)

        self.board.params["intamp_bias_mode_chan0_7"] = chan0_7
        self.board.params["intamp_bias_mode_chan8_15"] = chan8_15

    def _validate_inputs_or_raise(self, bias_mode):

        if bias_mode is not None:
            if not isinstance(bias_mode, str):
                raise TypeError("bias_mode must be a string")
            if bias_mode.lower()[0] not in ["c", "v"]:
                raise ValueError('Type must be either "current" or "voltage"')

    def _reset_side_selectors(self):
        for side in ["l", "r"]:
            self._write_digital_register(f"writemask_common_{side}", 0)

        # RegWrite wrapper for ease of testing

    def _write_analog_register(self, register, value):
        AnalogRegisters(self.board).write(register, value)

    def _write_digital_register(self, register, value):
        DigitalRegisters(self.board).write(register, value)



""" Pedestals controller

All pedestals management happens here.

Background:

    Pedestals are used to compensate for imperfections in the chip manufacturing
    process, by capturing enough events from the storage array it's possible to
    statistically rule out data originating from these imperfections rather
    than the real world event.

Instructions:

    This controller can run as a one-shot:
    PedestalsController(board, daq).generate_pedestals()

    or persistent:
    ped_ctrl = PedestalsController(board, daq)
    ped_ctrl.generate_pedestals()

.. todo::
    - _capture_data_for_pedestals currently works different for SiREAD, should be uniform.
    - Need a way to track progress and run non-blocking

"""
import copy
import gzip
import itertools
import logging
import os
import pickle
from typing import Iterable, Iterator

import numpy as np

from naludaq.backend.managers import AcquisitionManager
from naludaq.backend.models.acquisition import RemoteAcquisition
from naludaq.board import Board
from naludaq.communication import ControlRegisters, DigitalRegisters
from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.helpers.exceptions import OperationCanceledError, PedestalsDataCaptureError
from naludaq.helpers.helper_functions import event_transfer_time
from naludaq.tools.metadata import Metadata

LOGGER = logging.getLogger(
    "naludaq.pedestals_controller_new"
)  # pylint: disable=invalid-name


class PedestalsGeneratorNew:
    """Pedestals controller manages the pedestals generation.

    The pedestals are used to reduce noise originating from the chip itself.
    By capturing data from the board it's possible to use the mean value of x events to
    generate an average error for every sample. Since this average is
    an effect of the imperfections in the chip, subtracting the average counteracts
    the effect of the imperfections.

    The average is improving with 1/sqrt(n).

    This controller relies on an external data acquisiton capturing data to the
    .pedstore buffer it also relies on a board with an active connection to
    communicate with the hardware.

    The pedestals can then be used by the parser to remove the noise.

    Args:
        board:
        num_captures (int): Number of datapoints per sample, used for averaging the values.
        num_warmup_events (int): Number of initial events to discard. Helps the board
            settle and excludes events that may skew the pedestals.
        channels (list[int]): list of channels to generate pedestals for. Any channels excluded
            will result as `np.nan` in the generated pedestals.

    Attributes:
        num_captures: How many datapoints/samplepoint used to calculate the average.

    Raises:
        NotImplementedError if the given board does not support pedestals.
    """

    def __init__(
        self,
        board: Board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels=None,
    ):
        if not board.is_feature_enabled("pedestals"):
            raise NotImplementedError(
                f'Board "{board.model}" does not have support for pedestals.'
            )
        self._board = board

        self._progress = []
        self._cancel = False

        self._num_captures = num_captures
        self._num_warmup_events = num_warmup_events
        self._block_size = self.board.params["pedestals_blocks"]

        self._margin = 2
        self._sleep_time = 0
        self.channels = channels or range(self.board.channels)  # sets sleep time
        self._debug_raw_events = []

    @property
    def board(self):
        """Get/set board for pedestals capture.

        Raises:
            NotImplementedError if the given board does not support pedestals.
        """
        return self._board

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook the read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    @property
    def channels(self) -> list[int]:
        """Get/set the channels that will be read out. Any channels not enabled will result
        in NANs for that channel.
        """
        return self._channels

    @channels.setter
    def channels(self, channels: Iterable[int]):
        if not isinstance(channels, Iterable):
            raise TypeError("Channels must be a list[int]")
        if len(channels) == 0:
            raise ValueError("At least one channel must be given")
        if any(not isinstance(c, int) for c in channels):
            raise TypeError("Channels must be a list[int]")
        if any(not 0 <= c < self.board.channels for c in channels):
            raise ValueError("One or more channel numbers are out of bounds")
        self._channels = list(set(channels))

        # minimum sleep time changes with number of channels
        self._recompute_sleep_time()

    @property
    def block_size(self) -> int:
        """Get/set the number of windows per block"""
        return self._block_size

    @block_size.setter
    def block_size(self, block_size: int):
        self._block_size = block_size

    @property
    def metadata(self) -> Metadata:
        """Get a proxy object to use for accessing pedestals metadata."""
        return Metadata(self.board.pedestals)

    @property
    def raw_events(self) -> list[dict]:
        """List of events used to last generate the pedestals.
        
        These events are parsed, and otherwise not modified in any way.
        They can be used for analysis! 🥳🎉
        """
        return self._debug_raw_events

    def generate_pedestals(self):
        """Generates pedestals and stores them in the board.pedestals.

        Pedestals stored on the board, they will be used on the next acquisition.

        If canceled, the board is set back to its previous state, and no pedestals
        are generated.
        """
        self._cancel = False
        old_pedestals = self.board.pedestals
        self._debug_raw_events = []

        self._update_progress(0, "Saving board state")
        backup = self._backup_settings()

        self.reset_pedestals()
        self._store_board_metadata()

        try:
            self._generate_pedestals_metadata_pre()
            blocks = self._capture_data_for_pedestals()
            self._generate_pedestals_metadata_post()
        except (OperationCanceledError, KeyboardInterrupt):
            self.board.pedestals = old_pedestals
        except:
            self.board.pedestals = old_pedestals
            raise
        else:  # else runs before finally if no error is raised
            self._update_progress(80, "Processing data")
            self._store_raw_data(blocks)
            self._store_averaged_data()
        finally:
            self._update_progress(90, "Restoring board state")
            self._restore_backup_settings(backup)

    def cancel(self):
        """Cancels the pedestals generation as soon as possible.
        No pedestals are generated, and the board is restored to
        its previous state.

        Can only be called from a separate thread.
        """
        self._cancel = True

    def reset_pedestals(self):
        """Remove the current pedestal data.

        Replaces the pedestals with a set of blank pedestals. All values set
        to zero with the shape set from board params.
        """
        self.board.pedestals = self._create_empty_pedestals(self._num_captures)

    def _backup_settings(self) -> dict:
        """Backup settings that might get overwritten to a dict.

        Returns:
            dict with the backup settings:
                'readout_channels',
                'control_registers',
                'digital_registers'
        """
        backup = {
            "readout_channels": get_readout_controller(
                self.board
            ).get_readout_channels(),
            "control_registers": copy.deepcopy(
                self.board.registers.get("control_registers", {})
            ),
            "digital_registers": copy.deepcopy(
                self.board.registers.get("digital_registers", {})
            ),
        }

        return backup

    def _restore_backup_settings(self, backup_settings: dict):
        """Restore all backuped settings to the board.

        Returns:
            True if settings have been restored, False if no old settings were found.
        """
        if not backup_settings:
            return

        channels = backup_settings.get("readout_channels", None)
        if channels is not None:
            get_readout_controller(self.board).set_readout_channels(channels)

        # Restore backup
        for register_space in ["control_registers", "digital_registers"]:
            backup_registers = backup_settings.get(register_space, None)
            if backup_registers is not None:
                self.board.registers[register_space] = backup_registers
        DigitalRegisters(self.board).write_all()
        ControlRegisters(self.board).write_all()

    # ================================================================================
    # Capture
    # ================================================================================
    def _capture_data_for_pedestals(self) -> list[list[dict]]:
        """Capture raw data for pedestals.

        Returns:
            list[list[dict]]: list of data for blocks. Warmup events
                are removed from the output.

        Raises:
            PedestalsDataCaptureError: if pedestals failed to generate.
            OperationCanceledError: if pedestals generation was cancelled.
        """
        LOGGER.debug(
            "Capturing pedestals for %s. Block size=%s, channels=%s",
            self.board.model,
            self.block_size,
            self._channels,
        )
        self._write_readout_settings()

        total_windows = self.board.params["windows"]
        starting_windows = range(0, total_windows, self.block_size)
        num_blocks = len(starting_windows)
        required_events = self._num_captures + self._num_warmup_events
        blocks = []

        for block_idx, start_window in enumerate(starting_windows):
            LOGGER.info(f"Capturing data for block {block_idx}/{len(starting_windows)}")
            self._update_progress(
                10 + block_idx * 75 / num_blocks,
                f"Capturing block {block_idx+1}/{num_blocks}",
            )
            block = self._capture_block_or_raise(start_window, required_events)
            blocks.append(block[self._num_warmup_events :])
        return blocks

    def _capture_block_or_raise(self, start_window: int, captures: int) -> list[dict]:
        """Attempts to capture a data for the given block.

        Args:
            start_window (int): block start window
            captures (int): number of events in the block (captures + warmup)
            block_size (int): number of windows in the block

        Returns:
            list[dict]: list of events

        Raises:
            PedestalsDataCaptureError: if the necessary number of events could not be captured.
            OperationCanceledError: if pedestals generation was cancelled.
        """
        self._set_read_window(start_window)
        expected_block = self._calculate_expected_window_labels(
            start_window, self.block_size
        )
        self._start_readout()
        try:
            with AcquisitionManager(self.board).create_temporary() as acq:
                acq.set_output()
                return self._get_valid_events_or_raise(acq, expected_block, captures)
        except Exception as e:
            raise
        finally:
            self._stop_readout()

    def _get_valid_events_or_raise(
        self,
        acq: RemoteAcquisition,
        expected_block: list[int],
        count: int,
        attempts: int = 5,
    ) -> list[dict]:
        """Attempts to get validated events for the given block.

        Args:
            acq (RemoteAcquisition): the output acquisition to pull events from.
            expected_block (list[int]): the expected window labels
            count (int): number of events (captures + warmup)
            attempts (int): number of attempts. After each failed attempt,
                the sleep time will increase.

        Returns:
            list[dict]: list of validated events

        Raises:
            PedestalsDataCaptureError: if the necessary number of events could not be captured.
            OperationCanceledError: if pedestals generation was cancelled.
        """
        events = []
        for _ in range(attempts):
            needed_amount = count - len(events)
            stream = self._stream_valid_events(
                acq,
                start=-1,
                expected_block=expected_block,
            )
            try:
                events.extend(itertools.islice(stream, needed_amount))
                return events
            except TimeoutError:
                LOGGER.debug(
                    "Didn't get enough events (%s out of %s), increasing sleep time",
                    len(events),
                    count,
                )
                self._increase_sleeptime()
            except OperationCanceledError:
                raise
        msg = (
            "Failed to capture enough events. The board may be unresponsive "
            "and need to be power cycled/reinitialized."
        )
        raise PedestalsDataCaptureError(msg)

    def _stream_valid_events(
        self, acq: RemoteAcquisition, start: int, expected_block: list[int]
    ) -> Iterator[dict]:
        """A generator for validated events.

        Args:
            acq (RemoteAcquisition): the acquisition
            count (int): number of events to stream
            start (int): start event in the acquisition
            expected_block (list[int]): expected window labels
        """
        stream = acq.stream_parsed(
            timeout=self._sleep_time,
            start=start,
            skip_bad=True,
        )
        for event in stream:
            self._raise_if_canceled()
            if self._validate_event(event, expected_block):
                # debug events need to be a full copy, since
                # subclasses can do weird things with the events
                self._debug_raw_events.append(copy.deepcopy(event))
                yield event

    def _write_readout_settings(self):
        """Prepare the board for readout by writing necessary readout settings."""
        block_size = self.block_size
        windows = self.board.params["windows"]
        rc = get_readout_controller(self.board)
        try:
            rc.set_readout_channels(self._channels)
            rc.set_read_window(windows=block_size, write_after_trig=windows)
        except Exception as e:
            LOGGER.debug("Setting the windows failed due to %s", e)

    def _validate_event(self, event: dict, expected_block: list[int]) -> bool:
        """Returns true if the window labels matches the expected_block.

        The validation matches the received window_labels with the expected
        window labels to make sure the block contains only data from that
        block. The firmware buffer sometimes contains more data from a previous block
        this makes sure it doesn't enter the next block buffer.

        Args:
            event (dict): event to validate
            expected_block (list): a list of expected window numbers

        Returns:
            True if validated, False if the events windows doesn't match expected.
        """
        test_channel = self._channels[0]
        try:
            if np.all(event["window_labels"][test_channel] == expected_block):
                return True
        except Exception as error_msg:
            LOGGER.error(f"Event validation failed due to: {error_msg}")
            return False
        LOGGER.warning(
            "Expected: %s != returned: %s",
            expected_block,
            event["window_labels"][test_channel],
        )
        return False

    def _calculate_expected_window_labels(self, start_window, block_size) -> list[int]:
        """Calculate the expected window labels for a block.

        Args:
            start_window (int): start window of the block.
            block_size (int): Number of windows captured per block.

        Returns:
            List of expected window numbers.
        """
        windows = self.board.params["windows"]
        return np.arange(start_window, start_window + block_size) % windows

    # ================================================================================
    # Readout
    # ================================================================================
    def _start_readout(self):
        """Start a readout"""
        readout_settings = {
            "trig": "imm",
            "lb": "forced",
        }
        get_board_controller(self.board).start_readout(**readout_settings)

    def _stop_readout(self):
        """Stop the readout"""
        get_board_controller(self.board).stop_readout()

    def _set_read_window(self, start_window: int):
        """Set the read window from the given start window.

        Args:
            start_window (int): the start window of the block (lookback)
        """
        get_readout_controller(self.board).set_read_window(
            windows=self.block_size,
            lookback=start_window,
            write_after_trig=self.board.params["windows"],
        )

    # ================================================================================
    # Processing
    # ================================================================================
    def _create_empty_pedestals(self, num_captures: int):
        """Build the structure for raw pedestals data."""
        return {
            "data": np.full(
                shape=(
                    self.board.params["channels"],
                    self.board.params["windows"],
                    self.board.params["samples"],
                ),
                fill_value=np.nan,
            ),
            "rawdata": np.full(
                shape=(
                    self.board.params["channels"],
                    self.board.params["windows"],
                    self.board.params["samples"],
                    num_captures,
                ),
                fill_value=np.nan,
            ),
        }

    def _store_raw_data(self, blocks: list[list[dict]]):
        """Store raw data from blocks into the pedestals dict.

        Args:
            blocks (list[list[dict]]): blocks to store
        """
        block_size = self.block_size
        raw_data = self.board.pedestals["rawdata"]
        for block_idx, block in enumerate(blocks):
            for capture_number, event in enumerate(block):
                for chan in range(self.board.params["channels"]):
                    if chan not in self._channels:
                        continue
                    for window_num in range(block_size):
                        window = window_num + block_idx * block_size
                        # Avoid rolling over and overwrite block 0
                        if block_idx != 0 and (window < block_size):
                            continue
                        # avoid window number rolling over
                        if window >= self.board.params["windows"]:
                            continue
                        for sample in range(self.board.params["samples"]):
                            index = sample + window_num * self.board.params["samples"]
                            data = event["data"][chan][index]
                            raw_data[chan, window, sample][capture_number] = data

    def _store_averaged_data(self):
        """Generate processed pedestals data from the raw data."""
        raw_data = self.board.pedestals["rawdata"]
        self.board.pedestals["data"] = np.nanmean(raw_data, axis=3)

    # ================================================================================
    # Metadata
    # ================================================================================
    def _generate_pedestals_metadata_pre(self):
        """Adds some metadata to the pedestals dict. Called immediately
        before the pedestals data will be captured.
        """
        self._store_board_metadata()
        self._store_sensor_readings()

    def _generate_pedestals_metadata_post(self):
        """Adds some metadata to the pedestals dict. Called immediately
        after the pedestals data has been captured.
        """
        self._store_sensor_readings()

    def _store_board_metadata(self):
        """Store board params/registers into pedestals metadata"""
        self.metadata.set_configuration(self.board)

    def _store_sensor_readings(self):
        """Store sensor readings into pedestals metadata"""
        LOGGER.debug("Storing sensor metadata")
        self.metadata.store_sensor_readings(self.board)

    # ================================================================================
    # Import/Export
    # ================================================================================
    @staticmethod
    def save_pedestals(pedestals, filename):  # TODO(v.0.1.23): Use IO module.
        """Save the pedestal in binary format for backwards compatibility."""
        if not isinstance(pedestals, dict):
            raise TypeError(f"pedestals must be a dict, got {type(pedestals)}")
        if filename is None:
            raise TypeError("Supplied pathname is NoneType.")
        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        try:
            with gzip.GzipFile(filename, "wb", compresslevel=4) as f:
                pickle.dump(pedestals, f, protocol=pickle.HIGHEST_PROTOCOL)
        except IOError as error_msg:
            LOGGER.error("File could not be created. %s", error_msg)
        except pickle.PicklingError as error_msg:
            LOGGER.error("Saving pedestals failed: %s", error_msg)

    def load_pedestals(self, filename):  # TODO(v.0.1.23): Use IO module.
        """Load the pedestal gziped and pickled.

        The pedestals object is loaded to both self.pedestals
        and is returned.

        Args:
            filename: valid filename

        Returns:
            Pedestals data.
        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"No file found: {filename}")
        try:
            with gzip.GzipFile(filename, "rb") as f:
                outped = pickle.load(f)
        except EOFError as error_msg:
            LOGGER.error("Reached the end of the file, msg: %s", error_msg)
        except IOError as error_msg:
            LOGGER.error("File could not be loaded. %s", error_msg)
        except pickle.UnpicklingError as error_msg:
            LOGGER.error("Loading pedestals failed: %s", error_msg)
        else:
            if not isinstance(outped, dict):
                raise TypeError(f"Not a valid Pedestals file: {filename}")
            self.board.pedestals = outped
            return outped

    # ================================================================================
    # Helpers
    # ================================================================================
    def _recompute_sleep_time(self):
        """Recompute `self._sleeptime` using generation parameters."""
        self._sleep_time = event_transfer_time(
            self.board,
            windows=self.block_size,
            channels=len(self._channels),
            margin=self._margin,
            overhead=0.5,
        )
        LOGGER.debug("Waittime changed to %s", self._sleep_time)

    def _increase_sleeptime(self, current_amount=0, needed_amount=1):
        """Increase the sleeptime based on received vs expected.

        Formula is intended to change the margin incrementally without
        causing to much oscillation.
        """
        # Settles at 0.38
        factor = ((needed_amount - current_amount) / needed_amount) / 3 + 1
        self._margin = min(self._margin * factor, 5)
        self._recompute_sleep_time()

    def _raise_if_canceled(self):
        """Raise an ``OperationCanceledError`` if the cancel flag is set."""
        if self._cancel:
            raise OperationCanceledError("Pedestals generation was canceled.")

    def _update_progress(self, percent: float, message: str):
        """Updates a progress "receiver" with a percent and message.

        The receiver can be a list or deque of (percent, message) tuples,
        or a ProgressDialog (for use with NaluScope).

        Args:
            percent (float): the percent completion of the task
            message (str): a description of what is currently taking place
        """
        progress = self._progress
        if progress is None:
            LOGGER.debug("%s | %s", percent, message)
            return
        else:
            progress.append((percent, message))



from .context import Context
from .managers import (
    AcquisitionManager,
    BoardIoManager,
    ConfigManager,
    ConnectionManager,
)
from .models import (
    AvailableD2xxDevice,
    AvailableSerialDevice,
    D2xxDevice,
    Device,
    DeviceType,
    DiskAcquisition,
    DiskChunk,
    LocalAcquisition,
    RemoteAcquisition,
    SerialDevice,
    TemporaryAcquisition,
    UdpDevice,
)



from naludaq.backend.managers.base import Manager


class ConfigManager(Manager):
    def __init__(self, board):
        """Utility for higher-level configuration of the backend.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)

    def configure_packaging(self, events: "bytes | str", answers: "bytes | str"):
        """Sets the stop words used by the backend to separate
        events and answers.

        Args:
            events (bytes | str): stop word for events.
            answers (bytes | str): stop word for answers (register reads).
        """
        if isinstance(events, bytes):
            events = events.hex()
        if isinstance(answers, bytes):
            answers = answers.hex()
        try:
            self.context.client.put(
                "/server/data-format",
                params={
                    "model": self.board.model,
                    "events": events,
                    "answers": answers,
                },
            )
        except ValueError as e:
            raise ValueError("Stop words are invalid") from e



"""Class containing all functions to control the boards.

All functions to communicate with the Nalu boards are
collected here. The Facade naludaq is intended to simplify
the interface for general use.
"""
import logging

from .connection_controller import ConnectionController
from .upac import UPACConnectionController
from .upac96 import Upac96ConnectionController

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_connection_controller(board):
    """Get the correct connection controller based on the board model.

    Args:
        board (Board): the board object.

    Returns:
        The appropriate ConnectionController for the given board.
    """

    return {
        "upac32": UPACConnectionController,
        "upaci": UPACConnectionController,
        "upac96": Upac96ConnectionController,
        "zdigitizer": UPACConnectionController,
    }.get(board.model, ConnectionController)(board)



import inspect
from typing import Callable

from bayes_opt import BayesianOptimization, UtilityFunction

from naludaq.helpers.exceptions import OperationCanceledError


class BayesianOptimizer:
    """
    Wrapper function for the bayes_opt module, for any cost function or bounds, with support for
    progress, run, and cancel calls. Bayesian optimization is used for black-box functions where
    you want to find the maximum output by probing the input parameters, using Bayes Theorem to
    estimate the best next input parameters to try.

    References: https://distill.pub/2020/bayesian-optimization/
    """

    def __init__(self, board, cost_function, bounds, **kwargs):
        """
        Initialize a Bayesian Optimization object, which will try to find the maximum output of a
        cost function by probing within the given bounds. Discrete input space is not supported. If
        you want discrete inputs only, that conversion has to happen within the cost function.

        Args:
            board (Board): Board object with active connection
            cost_function (Callable): Function that returns a single float with inputs x and y
            bounds (dict): Dictionary with bounds of the parameter space.

        bounds example: {'x': [1100, 1500], 'y': [900, 1500]}
        """

        self.board = board
        self._cancel = False
        self._progress: list = []

        self.optimizer = BayesianOptimization(f=cost_function, pbounds=bounds, **kwargs)
        self.iteration_history = self.optimizer.res
        self._cancel_flag = 0
        self.iterations = 50

    def maximize(self, n_iter: int) -> dict:
        """
        Finds the input values that result in the maximum output value when run through the cost
        function.

        Args:
            n_iter (int): Number of iterations to run. (i.e. how many attempts to find max value.)

        Returns:
            dict: Dictionary with the stored input params that resulted in the largest output value.
        """
        utility = UtilityFunction(kind="ucb", kappa=2.5, xi=0.0)

        for i in range(n_iter):
            self._raise_if_canceled()

            next_point_to_probe = self.optimizer.suggest(utility)
            target = self.cost_function(**next_point_to_probe)

            self.optimizer.register(params=next_point_to_probe, target=target)

            self._progress.append((int(100 * i / n_iter), f"Scanning {(i+1)}/{n_iter}"))

        return self.optimizer.max

    def run(self):

        self.maximize(self.iterations)

    def cancel(self):

        self._cancel = True

    def _raise_if_canceled(self):
        """Raise an ``OperationCanceledError`` if the cancel flag is set."""
        if self._cancel:
            raise OperationCanceledError("Pedestals generation was canceled.")

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook to read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    @property
    def cost_function(self):

        return self._cost_function

    @cost_function.setter
    def cost_function(self, cost_func):
        if not type(Callable):
            raise TypeError("Cost Function has to be a function")
        self._verify_cost_function_or_raise(cost_func)
        self._cost_function = cost_func

    def _verify_cost_function_or_raise(self, cost_function):
        sig = inspect.signature(cost_function)

        try:
            x_param = sig.parameters["x"]
            y_param = sig.parameters["y"]
        except KeyError:
            raise KeyError("Cost function MUST have an x and y parameter")

        if x_param.annotation not in [float, int]:
            raise TypeError(
                "Cost function x parameter must be float or int (must have type signature to verify)"
            )

        if y_param.annotation not in [float, int]:
            raise TypeError(
                "Cost function y parameter must be float or int (must have type signature to verify)"
            )



import logging

from naludaq.backend.managers.io import BoardIoManager
from naludaq.communication import registers
from naludaq.helpers.exceptions import I2CError, InvalidBoardModelError
from naludaq.helpers.semiton import SemitonABC

logger = logging.getLogger("naludaq.i2c_registers")


class I2CRegisters(registers.ReadRegisters, registers.WriteRegisters, SemitonABC):
    def __init__(self, board):
        """Communication with I2C devices through FPGA registers.

        Args:
            board (Board): the board object.
        """
        if "i2c_registers" not in board.registers:
            raise InvalidBoardModelError(
                f'Board "{board.model}" is missing I2C registers or does not support I2C'
            )

        super().__init__(board, "i2c_registers")
        self._write_cmd = 0xAF
        self._read_cmd = 0xAD
        self.width_cmd = 2
        self.width_addr = 2  # hex char
        self.width_val = 4  # hex char

        # These are the same on all valid boards
        self._i2c_transmit_command = board.params.get("i2c", {}).get(
            "transmit_command", "CA000000"
        )

    def transmit_command(self):
        """Send the command which tells the board to transmit data over the I2C bus"""
        try:
            # Tell board to fire i2c commands
            if self.board.using_new_backend:
                BoardIoManager(self.board).write(self._i2c_transmit_command)
            else:
                self.board.connection.send(self._i2c_transmit_command)
        except Exception as e:
            raise I2CError("Failed to initiate I2C send") from e



from naludaq.parsers.headers.base import HeaderParser


class ASoCv2HeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """
        # head0: 1110TTTTTTTT
        # head1: TTTTTTTTTTTT
        # head2: TTTTWWWWWWWW
        trigger_time_ns = (raw_data[0] & 255) << 16  # 1
        trigger_time_ns += raw_data[1] << 4  # 2 # 24 bit number
        trigger_time_ns += raw_data[2] % 255
        event["prev_final_window"] = raw_data[2] & 255  # evt header 0
        event["trigger_time_ns"] = trigger_time_ns
        event["head0"] = raw_data[0]
        event["head1"] = raw_data[1]
        event["head2"] = raw_data[2]
        # event['event_id": raw_data[3] # 3 # not a thing, maybe will add later



"""A special data class that behaves like a list but pedestals correct on access."""

from collections import deque

from .pedestals_correcter import PedestalsCorrecter


class PedestalsAcq:
    def __init__(self, pedestals, acquisition=None):
        self.pedestals = pedestals
        self.pedestals_correcter = PedestalsCorrecter(pedestals)
        self.acquisition = acquisition

    def __getitem__(self, index):

        if isinstance(index, slice):
            # Get a slice from the deque
            # use itertools?
            # pedestals correct each
            ped_acq = deque()
            for ev in acq_slice:
                self.pedestals_correcter.run(ev)
            return ped_acq

        elif isinstance(index, int):
            event = self.acquisition[index]
            self.pedestals_correcter.run(event)
            return self.pedestals[index]



"""Peripherals controller controlls sensors on the board.

"""
from .peripherals_controller import PeripheralsController


def get_peripherals_controller(board) -> PeripheralsController:
    """Gets the appropriate peripherals controller for the given board.

    Since there is only one peripherals controller at the moment,
    the controller returned is always `PeripheralsController`.

    Args:
        board (Board): the board object

    Returns:
        The peripherals controller for the board.
    """
    if board.model in ["aodsoc_asoc", "aodsoc_aods"]:
        from .aodsoc import AodsocPeripheralsController

        return AodsocPeripheralsController(board)
    if board.model in ["hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]:
        from .hdsoc import HdsocPeripheralsController

        return HdsocPeripheralsController(board)
    if board.model in ["upac32"]:
        from .upac import UpacPeripheralsController

        return UpacPeripheralsController(board)
    if board.model in ["aardvarcv3"]:
        from .aardvarcv3 import Aardvarcv3PeripheralsController

        return Aardvarcv3PeripheralsController(board)

    from .peripherals_controller import PeripheralsController

    return PeripheralsController(board)



import abc

from naludaq.board.connections.base_connection import BaseConnection


class BaseSerialConnection(BaseConnection):
    @property
    @abc.abstractmethod
    def rtscts(self) -> bool:
        pass

    @rtscts.setter
    @abc.abstractmethod
    def rtscts(self, value: bool):
        self._set_rtscts(value)



"""AARDVARCv3
"""
import time
from collections import deque
from logging import getLogger

from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.tools.pedestals.pedestals_controller import PedestalsController, sleep_calc

LOGGER = getLogger("naludaq.aardvarcv3_pedestals_generator")


class AAv3PedestalsController(PedestalsController):
    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: "list[int]" = None,
    ):
        super().__init__(board, num_captures, num_warmup_events, channels=channels)
        self.margin: float = 2.0
        self.sleeptime: float = sleep_calc(board, self.margin, len(self._channels))

    def _capture_block(self, block, block_size, num_captures, num_warmup_events):
        num_captures = num_captures or self.num_captures
        num_warmup_events = num_warmup_events or self.num_warmup_events
        total_captures = num_captures + num_warmup_events

        self.validated_data = deque(maxlen=num_captures)
        get_readout_controller(self.board).set_read_window(lookback=block * block_size)

        t_count = 0
        timeout = 5

        while len(self.validated_data) < total_captures:
            # If not data is received after timeout, there is no recovery.
            if t_count > timeout:
                return False
            # Make sure buffer is empty
            self.block_buffer = deque(maxlen=total_captures + 1)
            self._daq.output_buffer = self.block_buffer

            needed_amount = total_captures - len(self.validated_data)
            if needed_amount <= 0:
                break

            # The capture logic, sw trigger and wait.
            self._start_capture()
            for _ in range(needed_amount + 1):
                get_board_controller(self.board).toggle_trigger()
                time.sleep(self.sleeptime)
            self._stop_capture()

            #  First window contains a bug, TODO: Fix me when bug in chip is fixed.
            try:
                _ = self.block_buffer.popleft()
            except IndexError:
                pass

            if len(self.block_buffer) != needed_amount:
                LOGGER.debug("data capture failed, restarting")
                self._increase_sleeptime(needed_amount)

            expected_block = self._calculate_expected_block(block, block_size)
            if not self.block_buffer:  # Rest of the logic require data.
                LOGGER.debug(
                    "Pedestals buffer is empty, indicating transmission error, retrying."
                )
                t_count += 1
                continue
            LOGGER.debug(
                "Only keep validated events, need: %s, from %s",
                needed_amount,
                len(self.block_buffer),
            )
            if self._only_keep_validated_events(
                self.block_buffer,
                self.validated_data,
                self._warmup_data if self._store_warmup_events else None,
                expected_block,
                num_captures,
                num_warmup_events,
            ):
                return True

        return True

    def _start_capture(self):
        """Start board and daq capture"""
        readout_settings = {
            "trig": "ext",
            "lb": "forced",
            "acq": "pedsub",
            "dig_head": False,
            "ped": "zero",
            "readoutEn": True,
            "singleEv": False,
        }
        self._daq.start_capture()
        get_board_controller(self.board).start_readout(**readout_settings)

    def _stop_capture(self):
        """Stop board and daq"""
        self._daq.stop_capture()
        get_board_controller(self.board).stop_readout()
        self._daq.stop_workers()



"""Initializer for the trbhm board.

It's based on the aardvarcv3 initializer because the trbhm
has two aardvarcv3 chips.
"""
import logging

from naludaq.board.initializers.aardvarcv3 import InitAardvarcv3

logger = logging.getLogger("naludaq.init_trbhm")


class InitTrbhm(InitAardvarcv3):
    def __init__(self, board):
        """Initializer for the TRBHM.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board)
        self.power_sequence = [
            "2v5_en",
            "1v2_en",
            "3v3_i2c_en",
            "clk2v5_en",
            "clk1v8_en",
            "clk_i2c_sel",
        ]

    def run(self) -> bool:
        """Runs the initialization sequence.

        Returns:
            True, always.
        """
        super().run()

        return True

    def _set_chip_tx_enabled(self, chips: list = [0, 1]):
        """Set whether the TX lines are enabled for each chip.
        Can be used to write a register on a single chip instead of both.
        """
        self.control_write("ard_tx_en", sum([1 << x for x in chips]))



from typing import List

from naludaq.communication import ControlRegisters
from naludaq.helpers.exceptions import RegisterFileError

from .base import BaseDACController


class DacControllerUpac32(BaseDACController):
    def __init__(self, board) -> None:
        """DAC controller for the UPAC32.

        Writes to control registers instead of through I2C/SPI
        to set the external DAC.

        Args:
            board (Board): the board object

        Raises:
            RegisterFileError: if the `ext_dac` field of the board
                params is invalid.
        """
        super().__init__(board)

    def _write_dacs(self, channels: List[int]):
        """Writes the DAC values held in board object.

        Args:
            channels (list[int]): a list of which DAC values to write. The list
                should contain elements from the `register_mapping` field in the params.
        """
        self._validate_channels(channels)

        for chan in channels:
            value = self.board.dac_values[chan]
            register = self.dac_params["register_mapping"][chan]
            self._write_control_register(register, value)

        self._write_control_register("dac_write_strobe", True)
        self._write_control_register("dac_write_strobe", False)

    def _validate_ext_dac_params(self, params: dict):
        """Validates the ext_dac board parameters.

        Runs the same checks as the DAC controller base, but also
        makes sure the `register_mapping` field is correct.

        Raises:
            RegisterFileError if the ext_dac params are invalid and cannot be used
        """
        super()._validate_ext_dac_params(params)

        dac_params = params["ext_dac"]
        channels = set(dac_params["channels"].keys())
        register_mapping = dac_params.get("register_mapping", None)

        if register_mapping is None:
            raise RegisterFileError("Ext DAC params are missing the register mapping")
        if not isinstance(register_mapping, dict):
            raise RegisterFileError("Register mapping for ext DAC must be a dict")
        if set(channels) != set(register_mapping.keys()):
            raise RegisterFileError(
                "Register mapping for ext DAC does not match the available DAC channels"
            )
        if any(not isinstance(x, str) for x in register_mapping.values()):
            raise RegisterFileError(
                "Register mapping values for ext DAC must be register names"
            )

    def _validate_channels(self, channels: List[int]):
        """Makes sure the channels to set DACs for are correct.

        Runs the same checks as the DAC controller base, but also
        makes sure that only channels defined in the board params
        are present.
        """
        super()._validate_channels(channels)
        if len(set(channels) - set(self.dac_params["channels"].keys())) != 0:
            raise ValueError(f"Incorrect channels given ({len(channels)})")

    def _write_control_register(self, name: str, value: int):
        """Writes a control register."""
        ControlRegisters(self.board).write(name, value)  # pragma: no cover



"""
"""
import logging

from naludaq.helpers.exceptions import InvalidTriggerMaskError

from .default import TriggerController

LOGGER = logging.getLogger("naludaq.trigger_controller_upac")


class TriggerControllerUpac(TriggerController):
    """TriggerController for UPAC series of boards.

    Based on the default Controller

    """

    def __init__(self, board):
        super().__init__(board)
        self._num_chips = board.params.get("num_chips", 4)
        self._num_trig_chans = board.params.get("num_trigger_channels", 8)

    def set_trigger_mask(self, chips: list):
        """Sets the chip trigger mask.

        Currently only supports UPAC32, UPAC-I, and Z-Digitizer boards.

        Args:
            chips (list): a list of enabled channels

        Raises:
            TypeError if the given list is not actually a list
            InvalidTriggerMaskError if the list supplied is the wrong length
                or contains invalid elements
        """
        self._validate_chip_list_or_raise(chips)

        reg_name = "trigger_mask_enable"
        trigmask = 0xF0  # Only the lower 4 bits are used
        for chip in chips:
            trigmask |= 1 << chip

        LOGGER.debug(f"Setting trigger mask: {reg_name}: {trigmask:04b}")
        self._write_control_register(reg_name, trigmask)

    def _validate_chip_list_or_raise(self, chips):
        """Validates a list of chips, and raises an error
        if there's a problem with them.

        Args:
            chips: the object that is supposedly a chip list

        Raises:
            TypeError if the chip list is not a list
            InvalidTriggerMaskError if the list is too long,
                contains elements that are not an int,
                or contains elements that are too large.
        """
        if not isinstance(chips, list):
            raise TypeError(f"Chips needs to be a list, not a {type(chips)}")
        elif len(chips) > self._num_chips:
            raise InvalidTriggerMaskError(f"Too many chips, max is {self._num_chips}")
        for chip in chips:
            if not isinstance(chip, int):
                raise InvalidTriggerMaskError(f"Chips can only contain integers")
            elif chip >= self._num_chips or chip < 0:
                raise InvalidTriggerMaskError(f"Chip number {chip} is out of bounds")

    def set_trigger_edge(self, raising: bool = True):
        """Set which signal edge to trigger on.

        Shift between positive going signals (raising) and negative (falling).

        Args:
            raising(bool): If true, trigger on positive going signals, else falling edge.

        Raises:
            TypeError if raises is not a bool.
        """
        if not isinstance(raising, bool):
            raise TypeError("raising must be a bool, got %s", type(raising))

        self._write_control_register("trigger_sign", raising)

    def _set_trigger_thresholds(self):
        """Sets the trigger values and send to hardware.

        Take the trigger values from the board
        """
        trigger_values = self.board.trigger_values

        for chan in range(self._num_trig_chans):
            register = f"psec4a_trigthresh{chan}_out"
            tval = trigger_values[chan]
            self._write_control_register(register, tval)

    def write_triggers(self):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        self._set_trigger_thresholds()



"""
"""
import multiprocessing as mp
import pathlib
import queue
import threading
import time
from datetime import datetime
from logging import getLogger

from naludaq.daq.preprocess import Preprocess
from naludaq.io.io_manager import IOManager

LOGGER = getLogger(__name__)


class Constants:
    # Cool hack to prevent serialization error
    def __init__(self, board):
        self.pedestals = (getattr(board, "pedestals", None),)
        self.params = board.params
        self.timingcal = getattr(board, "timingcal", None) or board.timingcal
        self.pedestals = getattr(board, "pedestals", None)
        self.caldata = getattr(board, "caldata", None)


class CSVStorageWorker:
    """Control class for the storage worker router and the worker."""

    def __init__(
        self,
        input_buffer,
        board,
        storage_settings,
        directory,
        namepattern="{}_event_{}",
    ):
        self.stopper = mp.Event()
        self.input_buffer = input_buffer
        self.board = board
        self.storage_settings = storage_settings
        self.intermediary_buffer = mp.Queue()
        self._directory = directory
        self._namepattern = namepattern
        self.router = None
        self.constants = Constants(board)
        self.router = deque2QueueRouter(
            self.input_buffer,
            self.intermediary_buffer,
        )
        self.worker = Queue2CSV(
            self.stopper,
            self.intermediary_buffer,
            self.constants,
            self.storage_settings,
            self.directory,
            self.namepattern,
        )

    def reset(self):
        """close the old and recreate the internal elements."""
        if self.router:
            self.router.stop()
        if self.intermediary_buffer:
            self.intermediary_buffer.close()
        self.intermediary_buffer = mp.Queue()
        self.router = deque2QueueRouter(
            self.input_buffer,
            self.intermediary_buffer,
        )
        self.worker = Queue2CSV(
            self.stopper,
            self.intermediary_buffer,
            self.board,
            self.storage_settings,
            self.directory,
            self.namepattern,
        )

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value
        message = ("directory", value)
        self.router.relay(message)
        # self.output_buffer.put(message)

    @property
    def namepattern(self):
        return self._namepattern

    @namepattern.setter
    def namepattern(self, value):
        message = ("pattern", value)
        self.router.relay(message)
        # self.output_buffer.put(message)
        self._namepattern = value

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, value):
        message = ("amount", value)
        self.router.relay(message)
        # self.output_buffer.put(message)
        self._amount = value

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        message = ("frequency", value)

        self.router.relay(message)
        # self.output_buffer.put(message)
        self._interval = 1 / value
        self._frequency = value

    def start(self):
        self.router.start()
        self.worker.start()

    def stop(self):
        self.router.stop()
        self.stopper.set()

    def buffer_level(self):
        """Returns a string wtih the buffer levels."""
        return f"{len(self.input_buffer)}, {self.intermediary_buffer.qsize()}"


class deque2QueueRouter(threading.Thread):
    """Communicates with the worker.

    All communication with the process is sent through this thread.


    """

    def __init__(self, input_buffer, output_buffer):
        super().__init__()
        self.amount = -1

        self.input_buffer = input_buffer
        self.output_buffer = output_buffer
        self._interval = 1 / 100
        self.running = True

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value
        message = ("directory", value)
        self.relay(message)

    @property
    def namepattern(self):
        return self._namepattern

    @namepattern.setter
    def namepattern(self, value):
        message = ("namepattern", value)
        self.relay(message)
        self._namepattern = value

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, value):
        # message = ("amount", value)
        # self.relay(message)
        self._amount = value

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        message = ("frequency", value)
        self.relay(message)
        self._interval = 1 / value
        self._frequency = value

    def relay(self, message):
        """Relay a message to the worker.

        Args:
            message (tuple): Sends a message (command(str), value(object)), must be pickleable.
        """
        if not isinstance(message, tuple):
            raise TypeError("Message must be a tuple")
        if not isinstance(message[0], str):
            raise ValueError("First part of message mst be a command.")
        try:
            self.output_buffer.put(message)
        except ValueError:
            raise ValueError("Communication path to worker is closed.")

    def run(self):
        """ """
        self._workloop()

    def stop(self):
        """ """
        self.output_buffer.put(("Stop", []))
        self.running = False

    def _workloop(self):
        """Run in thread, takes data from output_buffer and puts ina queue for CSV writer"""
        interval = self._interval  # 1/100 # 100Hz
        index = 0

        while self.running and (self.amount - index != 0):
            frametime = time.perf_counter()
            if index < len(self.input_buffer):
                evt = self.input_buffer[index]
                cmd = "save"
                index += 1
                message = (cmd, evt)
                self.output_buffer.put(message)

            try:
                time.sleep(frametime + interval - time.perf_counter())
            except ValueError:
                pass
        self.output_buffer.put(("stop", []))
        self.output_buffer.close()
        self.output_buffer.join_thread()


class Queue2CSV(mp.Process):
    """Worker storing the CSV files on disk from an input mp.Queue

    Also receives commands from the queue. By using the router object
    it'll route the messages to this process.
    """

    def __init__(
        self, stopper, input_buffer, constants, storage_settings, directory, namepattern
    ) -> None:
        super().__init__()
        self.daemon = True
        self.timeout = 100
        self.stopper = stopper  # mp.Event
        self.input_buffer = input_buffer  # mp.Queue
        self.preprocess = Preprocess(constants)
        self.io_manager = IOManager(constants.params)
        self.storage_settings = storage_settings
        self.directory = directory
        self.namepattern = namepattern
        self.filetype = ".csv"
        self.bad_filetype = ".bin"

    def run(self):
        self._worker_loop()

    def stop(self):
        self.running = False

    def _worker_loop(self):
        """Worker loop, runs until certain conditions are met."""
        self.running = True
        while not self.stopper.is_set() and self.running:
            try:
                message = self.input_buffer.get(block=True, timeout=self.timeout)
            except (queue.Empty, mp.TimeoutError):
                continue
            cmd = message[0]
            evt = message[1]
            if cmd.lower() == "save":
                self.parse_and_save(evt)
            elif cmd.lower() == "stop":
                self.running = False
                continue
            elif cmd.lower() == "kill":
                break
            elif cmd in ["directory", "namepattern", "frequency", "amount"]:
                try:
                    setattr(self, cmd, evt)
                except AttributeError:
                    # Not a valid command
                    pass

        # self.input_buffer.close()
        # self.input_buffer.join_thread()

    def parse_and_save(self, evt):
        """
        TODO: Modify this function with settings for filename and file path.
        """
        parsed_evt = preprocess_event(evt, self.preprocess, self.storage_settings)

        evt_num = evt.get("event_num", None)
        timestamp = evt.get("created_at", None)
        if timestamp is None:
            timestamp = datetime.now().strftime("%y-%m-%d_%H-%M-%S")
        else:
            timestamp = datetime.fromtimestamp(timestamp)
            timestamp = timestamp.strftime("%y-%m-%d_%H-%M-%S")
        if evt_num is not None:
            filename = self.namepattern.format(timestamp, evt_num)
            filepath = pathlib.Path(self.directory) / (filename + self.filetype)
            bad_filepath = pathlib.Path(self.directory) / (filename + self.bad_filetype)
        else:
            filepath = unique_path(self.directory, self.namepattern)
            bad_filepath = unique_path(self.directory, self.namepattern)
        if parsed_evt.get("data", None) is not None:
            self.io_manager.export_acq_to_csv(filepath, [parsed_evt])
        else:
            self.io_manager.export_acq_to_bin(bad_filepath, [evt])


def preprocess_event(event, preprocess, storage_settings):
    """Create a parsed and pedestals corrected event from raw input.

    Args:
        event (dict): raw event
        preprocess (Preprocess): Processor to convert raw -> parsed events
        storage_settings (dict): Settings for how preprocess will parse events

    Returns:
        event dictionary with parsed and pedestals corrected event.
    """
    if event.get("data", None) is None:
        try:
            event = preprocess.run(
                event,
                **storage_settings,
            )
        except Exception:
            event = {}
            # raise

    return event


def unique_path(directory, name_pattern):
    counter = 0
    while True:
        counter += 1
        path = directory / name_pattern.format(counter)
        if not path.exists():
            return path



import logging
from typing import Iterator

from naludaq.backend.models.acquisition import RemoteAcquisition
from naludaq.controllers.board import get_board_controller

from .default import PedestalsGeneratorNew

LOGGER = logging.getLogger(
    "naludaq.pedestals_controller_aav3_new"
)  # pylint: disable=invalid-name


class PedestalsGeneratorAardvarcv3New(PedestalsGeneratorNew):
    def _start_readout(self):
        """Start a readout"""
        readout_settings = {
            "trig": "ext",
            "lb": "forced",
            "acq": "pedsub",
            "dig_head": False,
            "ped": "zero",
            "readoutEn": True,
            "singleEv": False,
        }
        get_board_controller(self.board).start_readout(**readout_settings)

    def _stream_valid_events(
        self,
        acq: RemoteAcquisition,
        start: int,
        expected_block: list[int],
    ) -> Iterator[dict]:
        """A generator for validated events.

        Args:
            acq (RemoteAcquisition): the acquisition
            count (int): number of events to stream
            start (int): start event in the acquisition
            expected_block (list[int]): expected window labels
        """
        bc = get_board_controller(self.board)
        bc.toggle_trigger()
        for event in super()._stream_valid_events(acq, start, expected_block):
            yield event
            bc.toggle_trigger()



"""Tools to help explore FT600 connected devices
"""
import functools
from contextlib import contextmanager

from naludaq.board.connections.connection_factory import USB3_AVAILABLE
from naludaq.helpers.exceptions import FTDIError

if USB3_AVAILABLE:
    import ftd3xx as ftd
    import ftd3xx.defines


def is_d3xx_available() -> bool:
    """Check if FTDI is available. This module can still be loaded without FTDI,
    so use this function if it is necessary to test the presence of drivers.
    """
    return USB3_AVAILABLE


def requires_ftd3xx(func):
    """Decorator. Use on functions to raise an FTDIError when FTDI is not present."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not USB3_AVAILABLE:
            raise FTDIError(
                "FTDI is not available because the drivers could not be found or loaded."
            )
        return func(*args, **kwargs)

    return wrapper


@requires_ftd3xx
def list_ft60x_devices() -> dict:
    """List all available ft60x devices.

    Returns:
        dict: the available devices formatted as {index: {<info>}}
    """
    ftd.createDeviceInfoList()
    devices = ftd.getDeviceInfoList() or {}
    devices = {
        i: {
            "id": device.ID,
            "serial": device.SerialNumber.decode(),
            "description": device.Description.decode(),
            "flags": device.Flags,
            "log_id": device.LocId,
            "is_open": (device.Flags & ftd3xx.defines.FT_FLAGS_OPENED) != 0,
        }
        for i, device in enumerate(devices)
    }
    return devices


@contextmanager
def connect(*args, **kwds):
    connection = _get_connection(*args, **kwds)
    try:
        yield connection
    finally:
        connection.close()


def _get_connection(index):
    devices = ftd.getDeviceInfoList()
    connection = ftd.create(devices[index].ID)
    return connection


@requires_ftd3xx
def serial_from_index(index):
    """Return the serialnumber from the index number"""

    connection = _get_connection(index)
    if connection is None:
        raise ConnectionError(f"Can't connect to hardware at index: {index}.")
    cfg = connection.getChipConfiguration()
    connection.close()

    desc = bytearray(cfg.StringDescriptors)
    d = desc[desc[0] + desc[desc[0]] + 2 :]

    ser_no = ""
    for c in range(2, len(d), 2):
        if d[c] == 0:
            continue
        ser_no += "{0:c}".format(d[c])

    return ser_no


@requires_ftd3xx
def index_from_serial(serial_no: str) -> int:
    """Return the serial number for an index"""
    devices = ftd.getDeviceInfoList()
    if len(devices) == 0:
        raise ConnectionError("Can't find any connected hardware.")
    for device in devices:
        idx = device.ID
        try:
            serno = serial_from_index(idx)
        except ConnectionError:
            continue
        if len(serial_no) > len(serno):
            raise ValueError("Serial number is longer than board serial number")

        if serno[-len(serial_no) :] == serial_no:
            return idx
    raise ConnectionError("No board with that serial number found.")



def get_pedestals_controller(
    board,
    num_captures: int = 10,
    num_warmup_events: int = 10,
    channels: "list[int]" = None,
):
    """Gets the appropriate pedestals controller for a board.

    Args:
        board (Board): the board
        num_captures (int): Number of datapoints per sample, used for averaging the values.
        num_warmup_events (int): Number of initial events to discard. Helps the board
            settle and excludes events that may skew the pedestals.
        channels (list[int]): channels to generate pedestals for. Defaults to all channels.

    Returns:
        The pedestals controller

    Raises:
        NotImplementedError if the given board does not support pedestals.
    """
    if not board.is_feature_enabled("pedestals"):
        raise NotImplementedError(
            f'Board "{board.model}" does not have support for pedestals.'
        )
    if board.using_new_backend:
        from ._new import get_pedestals_generator_new

        return get_pedestals_generator_new(
            board, num_captures, num_warmup_events, channels
        )

    if board.model in ["upac32", "upaci", "zdigitizer"]:
        from .upac32_pedestals_controller import UpacPedestalsController

        return UpacPedestalsController(board, num_captures, num_warmup_events, channels)
    elif board.model in ["aardvarcv3", "aardvarcv4", "trbhm", "aodsoc_aods", "aodsoc_asoc"]:
        from .aardvarcv3_pedestals_generator import AAv3PedestalsController

        return AAv3PedestalsController(board, num_captures, num_warmup_events, channels)
    elif board.model in ["udc16"]:
        from .udc16_pedestals_generator import UDC16PedestalsGenerator

        return UDC16PedestalsGenerator(board, num_captures, num_warmup_events, channels)
    elif board.model in ["upac96"]:
        from .upac96_pedestals_generator import Upac96PedestalsGenerator

        return Upac96PedestalsGenerator(
            board, num_captures, num_warmup_events, channels
        )
    elif board.model in ["hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]:
        from .hdsoc_pedestals_controller import HDSoCPedestalsGenerator

        return HDSoCPedestalsGenerator(board, num_captures, num_warmup_events, channels)
    else:
        from .pedestals_controller import PedestalsController

        return PedestalsController(board, num_captures, num_warmup_events, channels)



from logging import getLogger

from naludaq.controllers.board import get_board_controller

from . import ConnectionController

LOGGER = getLogger("naludaq.connection_controller_upac96")


class Upac96ConnectionController(ConnectionController):
    """Connection controller for UPAC96.

    UPAC96 boards have different, mutually exclusive paths for UART/FTDI and USB3.
    This connection controller attempts to set the correct FIFO on the board
    for the board's connection type.
    """

    def _get_sync_response(self) -> int:
        """Tries to read the identifier register."""
        self._try_switch_output_fifo()
        return super()._get_sync_response()

    def _try_switch_output_fifo(self):
        """Tries to switch the output FIFO on the board according to the
        board connection type.

        This will always work for switching from UART/FTDI to USB3, but may or
        may not work for UART/FTDI depending on whether the baud rate is matched.

        Raises:
            NotImplementedError: if the connection type is invalid.
        """
        conn_type = self.board.connection_info.get("type", None)
        if conn_type in ["uart", "ftdi"]:
            get_board_controller(self.board).enable_uart()
        elif conn_type in ["ft60x"]:
            get_board_controller(self.board).enable_usb()
        else:
            raise NotImplementedError(
                f"Invalid connection type for UPAC96: {conn_type}"
            )



from naludaq.parsers.headers.base import HeaderParser


class ASoCHeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """
        event["prev_final_window"] = raw_data[0] & 255  # evt header 0
        trigger_time_ns = raw_data[1] << 12  # 1
        trigger_time_ns += raw_data[2]  # 2 # 24 bit number
        event["trigger_time_ns"] = trigger_time_ns
        event["event_id"] = raw_data[3]  # 3
